function frequenciesPerHorizontalPixel(spectrum, width) {
	return parseInt(spectrum.length / width);
}

function scaleSpectrum(spectrum, width) {
	var freqsPerHorizontalPixel = frequenciesPerHorizontalPixel(spectrum, width);
	var resultSpectrum = new Array();

	for (var i = 0; i < spectrum.length - 2 * freqsPerHorizontalPixel; i += freqsPerHorizontalPixel) {
		var avg = 0;
		for (var j = 0; j < freqsPerHorizontalPixel; ++j) {
			avg = (parseFloat(spectrum[j + i]) + j * avg) / (j + 1);
		}

		resultSpectrum.push(avg);
	}

	while (resultSpectrum.length < width) {
		resultSpectrum.push(0.0);
	}

	return resultSpectrum;
}

function addEvent(obj, name, func) {
    if (obj.attachEvent) obj.attachEvent("on" + name, func);
	else obj.addEventListener(name, func, false);
}

function detectIE() {
    var ua = window.navigator.userAgent;
    var msie = ua.indexOf('MSIE ');
    var trident = ua.indexOf('Trident/');

    if (msie > 0) {
		// IE 10 or older => return version number
        return parseInt(ua.substring(msie + 5, ua.indexOf('.', msie)), 10);
    }

    if (trident > 0) {
        // IE 11 (or newer) => return version number
        var rv = ua.indexOf('rv:');
        return parseInt(ua.substring(rv + 3, ua.indexOf('.', rv)), 10);
    }

    // other browser
    return false;
}

var panelMode = {
	SIMPLE: "simple",
	AREASET: "areaset",
	BRUSH: "brush",
	STATUS: "status"
};

(function($) {

	$.widget("dvr.camerapanel", {
		
		/*
		Eventit:
			'click' - lähetetään, kun canvasta klikataan
			'drag'
			'dragend'
			'maskchanged'
			'areaschanged'
			'areaselected'
		*/

		options: {
			camera: {
				label: "",
                uuid: ""
			},
			uid: "",
			width: 0,
			height: 0,
			rtspMode: true,
			rtspPort: 0,
			viewId: 0,
			windowless: true, // Jos true, ei rautakiihdytystä, mutta voidaan piirtää videoelementin päälle.
			networkCaching: 100,
			doubleClickFullscreen: true, // Huom! Tämä ignorataan, jos ei olla rtsp-moodissa ja click-signaali lähetetään aina vaikka tämä olisikin tosi
			mode: panelMode.SIMPLE,
			selectedArea: "",
			error: "",
			flipv: false,
			fliph: false,
			rotation: 0,
			mask: "",
			flipChanged: false,
			rotationChanged: false
		},
		
		_setOption: function(key, value) {
			if (key === "mode") {
				this.options.mode = value.toLowerCase();
				this.clearStatus();
				if (this.options.mode === panelMode.BRUSH) this._showBrushPoints(this._mouseOver);
			} else if (key === "error") {
				this.options.error = value;
				this._draw();
			} else if(key === "flipv") {
				this.options.flipv = value != "0";
				this.options.flipChanged = true;
			} else if(key === "fliph") {
				this.options.fliph = value != "0";
				this.options.flipChanged = true;
			} else if(key == "rotation") {
				this.options.rotation = parseInt(value);
				this.options.rotationChanged = true;
			}
		},
		
		_url: function (rtspPort, viewId, width, height, uid, camera) {
			return "rtsp://" + window.location.hostname + ":" + rtspPort + "/view/" + viewId + "?width=" + width + "&height=" + height + "&streamId=" + uid + "&camera=" + camera;
		},
		
		_fbvlcObject: function (windowless, width, height, target) {
			if (this.fbvlc) {
				return $("<object class='camerapanel-area' style='position: absolute;'>"
					+ "<param name='windowless' value='" + (windowless ? "true" : "false") + "' />"
					+ "<param name='toolbar' value='false' />"
					+ "<param name='fullscreen-toolbar' value='false' />"
					+ "<param name='network-caching' value='" + this.options.networkCaching + "' />"
					+ "<param name='bgcolor' value='#4d5563' />"
					+ "<param name='branding' value='false' />"
					+ (target ? "<param name='target' value='" + target + "' />" : "")
					//+ "<param name='native-scaling' value='true' />" 
					+ "</object>")
					.attr("type", "application/x-fb-vlc").attr("width", width).attr("height", height);
			} else {
				return $("<embed class='camerapanel-area' style='position: absolute;'></embed>")
					.attr("type", "application/x-vlc-plugin").attr("pluginspage", "http://www.videolan.org").attr("width", width).attr("height", height)
					.attr("windowless", (windowless ? "true" : "false")).attr("controls", "false").attr("bgcolor", "#4d5563").attr("branding", "false");
			}
		},
		
		play: function () {
			if (!this.options.rtspMode || this.options.rtspPort === 0) return;

            var url = this._url(this.options.rtspPort, this.options.viewId, this.options.width, this.options.height, this.options.uid, this.options.camera.uuid);
			
			if (this.fbvlc) {
				this.video[0].play(url);
			} else {
				if (!this.video[0].playlist) return;

				this.video[0].playlist.items.clear();
				this.video[0].playlist.add(url, "", ":network-caching=" + this.options.networkCaching);
				this.video[0].playlist.play();
			}
		},
		
		stop: function () {
			if (!this.options.rtspMode) return;
			
			if (this.fbvlc) this.video[0].stop();
			else {
				if (this.video[0].playlist) this.video[0].playlist.stop();
			}
		},

		_create: function () {
			this.fbvlc = detectIE();

			this.create();
		},
		
		isSimpleMode: function () { return this.options.mode === panelMode.SIMPLE; },
		isAreaSetMode: function () { return this.options.mode === panelMode.AREASET; },
		isBrushMode: function () { return this.options.mode === panelMode.BRUSH; },
		isStatusMode: function () { return this.options.mode === panelMode.STATUS; },
		
		_shouldShowStatus: function () {
			return this.isStatusMode() || (this.isBrushMode() && !this._mouseOver);
		},
		_shouldShowAudioStatus: function () {
			return this._shouldShowStatus();
		},
		
		// Palauttaa globaalin koordinaatin kameran kuva-alueen koordinaatistossa
		globalToFrame: function (x, y) {
			var offset = $(this.renderer.domElement).offset();
			return new THREE.Vector2(x - offset.left, y - offset.top);
		},
		
		// Palauttaa canvasin koordinaatin muutettuna scenen (eli three.js:n scenen) koordinaatiksi
		frameToScene: function (x, y) {
			 return new THREE.Vector2(x - this.renderer.domElement.width * 0.5, -y + this.renderer.domElement.height * 0.5);
		},

		// Palauttaa globaalin koordinaatin normalisoituna välille -1 ja 1.
		globalToNDC: function (x, y) {
			var p = this.globalToFrame(x, y);
            return this.frameToNDC(p.x, p.y);
		},
		
		frameToNDC: function (x, y) {
			return new THREE.Vector2((x / this.renderer.domElement.width) * 2 - 1, -(y / this.renderer.domElement.height) * 2 + 1);
		},

		create: function () {
			this.options.uid = 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
				var r = Math.random()*16|0, v = c == 'x' ? r : (r&0x3|0x8);
				return v.toString(16);
			});
			
			var that = this;
			
			this.wormDirection = 2;
			this.theGameIsOn = false;

			this.video = null;
			this.canvas = null;
			this.image = null;
			this.status = "";
			this.recordingStatus = false;
			this._mouseOver = false;
			
			this.tmpImage = new Image();
			this.imageUpdated = false;
			this.tmpImage.onload = function() {
				that.image = that.tmpImage;
				that.imageUpdated = true;
			}
			
			this.errorImage = new Image();
			this.errorImage.src = "images/no_signal_42px.png";
		
			this.fontSize = 12;
			this.font = this.fontSize + "px Arial";
			
			this.textColor = "rgba(255, 255, 255, 0.6)";
			this.backgroundColor = "rgba(0, 0, 0, 0.8)";
			this.defaultMargin = 5;
			
			this.container = $("<div></div>").css("position", "relative")
			.css("width", this.options.width)
			.css("height", this.options.height)
			.css("background-color", "#000000");

			if (this.options.rtspMode) {
				this.video = this._fbvlcObject(this.options.windowless, this.options.width, this.options.height);
				this.container.append(this.video);				
				if (!this.options.windowless) { // Window-tyyppisen päälle ei voida lisätä canvasta
					$(this.element).html(this.container);
					return;
				}
			}

			this._areas = []; // Lista alueita kuvaavista pistejoukoista
			this.areaObjects = []; // Lista alueiden graafiseen esittämiseen tarvittavista objekteista

			this.scene = new THREE.Scene();
 
			this.camera = new THREE.OrthographicCamera(this.options.width / -2, this.options.width /2, 
				this.options.height /2, this.options.height / -2, 0.0, 10);

			// Kameran framen esittämiseen liityvät
			this.frameGeometry = new THREE.BoxGeometry(this.options.width, this.options.height, 0);
			this.frameMaterial = new THREE.MeshLambertMaterial();
			this.frameMesh = new THREE.Mesh(this.frameGeometry, this.frameMaterial);
			this.scene.add(this.frameMesh);
			
			var ambientLight = new THREE.AmbientLight(0xffffff);
			this.scene.add(ambientLight);
			
            this.renderer = new THREE.WebGLRenderer( { antialias: false, alpha: true } );
			this.renderer.setSize(this.options.width, this.options.height);
			$(this.renderer.domElement).css("position", "absolute")
			this.container.append(this.renderer.domElement);

            this.brushPointObjects = [];
            this._updateBrushSize();
			this.setMaskColor("#3367d5")
			
			this.motionPointObjects = [];
			this._updateMotionPointSize();

			this._animate();

			// Tehdään canvas-elementti
			this.canvas = $("<canvas tabindex='1' class='camerapanel-area' style='position: absolute; left: 0; outline: none'></canvas>").attr("width", this.options.width).attr("height", this.options.height);
			this.context = this.canvas[0].getContext("2d");
			this.container.append(this.canvas);
			$(this.element).html(this.container);
			
			var dragging = false;
			var raycaster = new THREE.Raycaster();
		
			$(this.canvas).on("click", function (event) {		
				if (!that.isEnabled()) return;

				that.dragging = false;
				that._trigger("click");
				
				if (that.isAreaSetMode()) {
					// Haetaan minkä alueen tai kärjen alueella ollaan klikattu
					raycaster.setFromCamera(that.globalToNDC(event.pageX, event.pageY), that.camera);
					for (var i = 0; i < that.areaObjects.length; ++i) {
						var area = that.areaObjects[i];
						var intersects = raycaster.intersectObjects(area["areaEdges"]);
						if (intersects.length === 1) {
							var object = intersects[0].object;
							var index = area["areaEdges"].indexOf(object);
							// Haetaan tuplaklikatun viivan keskipiste..			
							var center = new THREE.Line3(object.geometry.vertices[0], object.geometry.vertices[1]).center();
							// ..ja lasketaan sen suhteellinen sijainti
							var relCenter = that._absToRel(center);
							
							// Haetaan ne aluepisteet joiden perusteella alue oli piirretty..
							var area = that._areas[i]["points"]; // areaObjects-listan alkio i on tehty areas-listan alkion i perusteella
							area.splice(index + 1, 0, relCenter); // ..ja lisätään uusi piste
							that._updateAreas();
							return;
						}
					}
					return;
				}
				
				if (!that.options.doubleClickFullscreen || !that.options.rtspMode || !that.fbvlc) return;

                var url = that._url(that.options.rtspPort, that.options.viewId, window.screen.availWidth, window.screen.availHeight, that.options.uid, that.options.camera.uuid);
				var fullscreenVideo = that._fbvlcObject(false, 100, 100, url);
				$(that.element).html(fullscreenVideo);
				
				addEvent(fullscreenVideo[0], "MediaPlayerPlaying", function () {
					// Kokoruututilan asettaminen toimii vasta, kun toisto on käynnistetty
					fullscreenVideo[0].fullscreen = true;
					
					function checkFullscreenExited() {
						if (fullscreenVideo[0].fullscreen) setTimeout(checkFullscreenExited, 100);
						else {
							$(that.element).html("");
							that.create();
							that.play(that.rtspPort, that.viewId);
						}
					};
					// Aletaan tarkastaa tietyin väliajoin ollaanko poistuttu kokoruututilasta
					checkFullscreenExited();
				});
			});

			var mouseX = 0;
			var mouseY = 0;
			var draggedArea = -1;
			var draggedNode = -1;
			var lastMoveTime = 0;
			var maskChanged = false;
			
			function doPTZ() {
				if (!dragging) {
					that._trigger("dragend");
					return;
				}

				var r = 100;

				var xy = that.globalToFrame(mouseX, mouseY);
				var xDiff = parseInt(xy.x * 2 * r / that.frameMesh.geometry.parameters.width - r);
				var yDiff = parseInt(xy.y * 2 * r / that.frameMesh.geometry.parameters.height - r);
				that._trigger("drag", null, {x: xDiff, y: yDiff, z: 0});
				
				setTimeout(doPTZ, 300);
			};
			
			$(this.canvas).on("keydown", function (event) {
				if (that.isAreaSetMode()) {
					if (event.which === 68) { // d
						if (that.options.selectedArea.length !== 0) that.removeArea(that.options.selectedArea);
						event.preventDefault();
					} else if (event.which === 65) { // a
						that._createArea("closed", [[0.1, 0.1], [0.3, 0.1], [0.15, 0.3]]);
					} else if (event.which === 76) { // l
						that._createArea("line", [[0.2, 0.2], [0.3, 0.5]]);
					}
				} else {
					if (event.which >= 37 && event.which <= 40) {
						that.wormDirection = event.which - 37;
						event.preventDefault();
					} else if (event.which === 87) {
						if (that.theGameIsOn) that.stopGame();
						else that.startGame();
						event.preventDefault();
					}
				}
			});

			$(this.canvas).on("mousedown", function (event) {
			
				if (!that.isEnabled()) return;

				event.preventDefault();
				dragging = true;
				mouseX = event.pageX;
				mouseY = event.pageY;

				if (that.isAreaSetMode()) {
					draggedNode = -1;
					draggedArea = -1;

					// Haetaan minkä alueen tai kärjen alueella ollaan klikattu
					raycaster.setFromCamera(that.globalToNDC(mouseX, mouseY), that.camera);
					
					for (var i = 0; i < that.areaObjects.length; ++i) {
						var area = that.areaObjects[i];
						// Tarkastetaan ensin klikattiinko jonkun kärjen päällä
						var intersects = raycaster.intersectObjects(area["nodes"]);
						if (intersects.length > 0) {
							draggedArea = i;
							draggedNode = area["nodes"].indexOf(intersects[0].object);
						} else if (area["type"] === "closed") {
							// Ei klikattu kärkeä, tarkastetaan klikattiinko aluetta
							intersects = raycaster.intersectObject(area["area"]);
							if (intersects.length > 0) draggedArea = i;
						} else if (area["type"] === "line") {
							intersects = raycaster.intersectObjects(area["areaEdges"]);
							if (intersects.length > 0) draggedArea = i;
						}
						
						if (draggedArea !== -1) {
							var selectedAreaChanged = (that.options.selectedArea != that._areas[i]["id"]);
							that.options.selectedArea = that._areas[i]["id"];
							that._updateAreas();
							if (selectedAreaChanged) that._trigger("areaselected");
							return;
						}
					}
				} else if (that.isBrushMode()) {
					maskChanged = false;
				} else {
					doPTZ();
				}
			});
			function __maskIndexFromCoords(x, y) {
				var maskRows = 36;
				var maskCols = 44;
				if(that.options.rotation == 90) {
					var maskx = Math.round(y / that.brushHeight);		
					var masky = maskRows - 1 - Math.round(x / that.brushWidth);
				} else if(that.options.rotation == 180) {
					var maskx = maskCols - 1 - Math.round(x / that.brushWidth);		
					var masky = maskRows - 1 - Math.round(y / that.brushHeight);
				} else if(that.options.rotation == 270) {
					var maskx = maskCols - 1 - Math.round(y / that.brushHeight);		
					var masky = Math.round(x / that.brushWidth);
				} else {
					var maskx = Math.round(x / that.brushWidth);
					var masky = Math.round(y / that.brushHeight);
				}
				
				if(that.options.fliph) {
					maskx = maskCols - 1 - maskx;
				}
				if(that.options.flipv) {
					masky = maskRows - 1 - masky;
				}
				return masky * maskCols + maskx;
			}

			$(this.canvas).on("mousemove", function (event) {
				if (event.button !== 0 || !dragging) return;
				
				// Käsitellään hiiren liikkeitä vain 20 millisekuntin välein
				var t = new Date().getTime();
				if (Math.abs(t - lastMoveTime) < 20) return;
				lastMoveTime = t;

				if (that.isAreaSetMode()) {
					var deltaX = event.pageX - mouseX;
					var deltaY = event.pageY - mouseY;
					if (draggedArea != -1 ) {
						var areaPoints = that._areas[draggedArea]["points"];
						// Jos draggedNode on asetettu, siirretään vaan sitä
						if (draggedNode != -1) {
							areaPoints[draggedNode][0] += deltaX / that.options.width;
							areaPoints[draggedNode][1] += deltaY / that.options.height;
						} else {
							for (var i = 0; i < areaPoints.length; ++i) {
								areaPoints[i][0] += deltaX / that.options.width;
								areaPoints[i][1] += deltaY / that.options.height;
							}
						}
						that._updateAreas();
					}
				} else if (that.isBrushMode()) {
					var framePos = that.globalToFrame(event.pageX, event.pageY);
                    var pos = that._toBrushPosition(framePos);
					
					var points = new Array;
					if (event.shiftKey) {
						for (var c = -2; c < 3; ++c) {
							for (var r = -2; r < 3; ++r) {
								var p = new THREE.Vector2(pos.x + that.brushWidth * c, pos.y + that.brushHeight * r);
								points.push(p);
							}
						}
					} else {
						points.push(pos);
					}
					
					for (var k = 0; k < points.length; ++k) {
						pos = points[k];
						raycaster.setFromCamera(that.frameToNDC(pos.x + 1, pos.y + 1), that.camera);
						intersects = raycaster.intersectObjects(that.brushPointObjects);
						if (event.ctrlKey) {
							for (var i = 0; i < intersects.length; ++i) {
								var index = that.brushPointObjects.indexOf(intersects[i].object);
								if (index != -1) that.brushPointObjects.splice(index, 1);
								that.scene.remove(intersects[i].object);
							}
							if (intersects.length > 0) {
								var maskidx = __maskIndexFromCoords(pos.x, pos.y);
								that.options.mask = that.options.mask.substring(0, maskidx) + '0' + that.options.mask.substring(maskidx + 1);
								maskChanged = true;
							}
						} else if (intersects.length === 0) {
                            that.brushPointObjects.push(that._drawBrushPoint(pos));
							// Päivitetään maski	
							var maskidx = __maskIndexFromCoords(pos.x, pos.y);
							that.options.mask = that.options.mask.substring(0, maskidx) + '1' + that.options.mask.substring(maskidx + 1);
							maskChanged = true;
						}
					}
				}
				mouseX = event.pageX;
				mouseY = event.pageY;
			});

			$(this.canvas).on("mouseup", function () { 
				dragging = false;
				if (that.isAreaSetMode() && draggedArea != -1) {
					var area = that.areaObjects[draggedArea];
					if (area["type"] === "closed") area["area"].material.opacity = 0.3;
					that._trigger("areaschanged");
				} else if (that.isBrushMode()) {
					if (maskChanged) that._trigger("maskchanged");
				}
			});
			
			$(this.canvas).on("mouseenter", function () {
				that._mouseOver = true;
                if (that.isBrushMode()) that._draw();
			});
			
			$(this.canvas).on("mouseout", function () {
				dragging = false;
				that._mouseOver = false;
                if (that.isBrushMode()) {
                    that._draw();
                    if (maskChanged) that._trigger("maskchanged");
                }
			});

			var mouseWheelId = 0; 			
			$(this.canvas).on("mousewheel", function (event) {
				if (!that.isEnabled()) return;
				
				that._trigger("zoom", null, event.deltaY * 100);
				
				// Lähetetään zoomend-eventti, kun viimeisestä mousewheel-eventistä on kulunut 500ms
				clearTimeout(mouseWheelId);
				mouseWheelId = setTimeout(function () {
					that._trigger("zoomend", null);
				}, 500);
			});
		},
		
		uid: function () { return this.options.uid; },
		
		width: function () { return this.options.width; },
		height: function () { return this.options.height; },
		
		resize: function (width, height) {
			this.options.width = width;
			this.options.height = height;
			this._updateBrushSize();
			this._updateMotionPointSize();
			if (this.container) this.container.css("width", width).css("height", height);
			
			if (this.video) {
				this.video.attr("width", width).attr("height", height);
				this.stop();
				this.play();
			}
			
			if (this.canvas) {
				this.canvas.attr("width", width).attr("height", height);
				this._draw();
			}

			if (this.camera) {
				this.camera.left = width / -2;
				this.camera.right = width / 2;
				this.camera.top = height / 2;
				this.camera.bottom = height / -2;
				this.camera.updateProjectionMatrix();
			}
			
			if (this.renderer) this.renderer.setSize(width, height);
		},
		
		update: function (imageSrc) {
			if (this.options.rtspMode) return;

			if (imageSrc.length === 0) {
				this.image = undefined;
				this.imageUpdated = true;
				this._draw();
			} else {
				this.tmpImage.src = imageSrc;
			}
		},
		
		updateAudioStatus: function (status) {
			this.audioStatus = status;
			this._draw();
		},

		updateStatus: function (status) {
			this.status = status;
			this._draw();
		},
		
		updateRecordingStatus: function (status) {
			this.recordingStatus = status;
			this._draw();
		},
		
		clearStatus: function () {
			this.status = "";
			this.audioStatus = null;
			this._draw();
		},

		areas: function () {
			return this._areas;
		},
		
		selectedArea: function () {
			return this.options.selectedArea;
		},
		
		addArea: function (area) {
			// Ei varmaan mitään järkeä säilyttää erikseen suhteellisia pisteitä ja piirrettyjä
			this._areas.push(area);
			this._updateAreas();
		},
		
		_createArea: function (type, points) {
			var area = new Map();
			area["id"] = Math.floor((Math.random() * 100000) + 100);
			area["title"] = "";
			area["type"] = type;
			area["points"] = points;
			this.addArea(area);
			this._trigger("areacreated", null, area);
		},
		
		removeArea: function (areaId) {
			for (var i = 0; i < this._areas.length; ++i) {
				if (this._areas[i]["id"] === areaId) {
					if (this.options.selectedArea === areaId) this.options.selectedArea = "";
					this._areas.splice(i, 1);
					this._updateAreas();
					this._trigger("arearemoved", null, areaId);
					return;
				}
			}
		},

		camera: function () { return this.options.camera; },
		setCamera: function (camera) {
			this.options.camera = camera;
			this.image = null;
		},
		
		mask: function () {
			return this.options.mask;
		},
		__transformMaskCoords: function(x, y) {
			var colCount = 44;
			var rowCount = 36;
			
			if(this.options.rotation == 90) {
				var maskX = (rowCount - 1 - y);
				var maskY = x;
			} else if(this.options.rotation == 180) {
				var maskX = (colCount - 1 - x);
				var maskY = (rowCount - 1 - y);
			} else if(this.options.rotation == 270) {
				var maskX = y;
				var maskY = (colCount - 1 - x);
				var inCanvas = new THREE.Vector2(y * this.brushWidth, (colCount - 1 - x) * this.brushHeight);
			} else {
				var maskX = x;
				var maskY = y;
			}

			if(this.options.fliph) {
				maskX = colCount - 1 - maskX;
			}

			if(this.options.flipv) {
				maskY = rowCount - 1 - maskY;
			}
			return {
				x: maskX,
				y: maskY
			}
		},
		setMask: function (mask) {
			this._clearBrushPoints();
			var colCount = 44;
			var rowCount = 36;

			for(var j = 0; j < rowCount; j++) {
				for(var i = 0; i < colCount; i++) {
					var idx = j * colCount + i;
					if (parseInt(mask[idx]) !== 1) continue;

					var transformedPoint = this.__transformMaskCoords(i, j);

					var inCanvas = new THREE.Vector2(transformedPoint.x * this.brushWidth, transformedPoint.y * this.brushHeight);
					var point = this._drawBrushPoint(inCanvas);
					if (!this.isBrushMode()) point.visible = false;
					this.brushPointObjects.push(point);
				}
			}

			this.options.mask = mask;
		},
		setMaskColor: function (maskColor) {
			var m = this.mask();
			this.brushMaterial = new THREE.MeshBasicMaterial({ color: maskColor, transparent: true, opacity: 0.6 });
			this.setMask(m);
		},
		
		setMotion: function (motion) {
			this._clearMotionPoints();
			var colCount = 44;
			var rowCount = 36;
			for(var j = 0; j < rowCount; j++) {
				for(var i = 0; i < colCount; i++) {
					var value = parseInt(motion.substr(j * colCount * 3 + i * 3, 3));
					if (value == 0) continue;

					var transformedPoint = this.__transformMaskCoords(i, j);

					var inCanvas = new THREE.Vector2(transformedPoint.x * this.brushWidth, transformedPoint.y * this.brushHeight);
					var point = this._drawMotionPoint(inCanvas, value);
					this.motionPointObjects.push(point);
				}
			}
		},
		
		setRTSPport: function (rtspPort) { this.options.rtspPort = rtspPort; },
		
        isEnabled: function () { return this.options.camera.uuid.length > 0; },
		clear: function () {
			this.options.camera.label = "";
            this.options.camera.uuid = "";
			this.image = null;
			this.clearStatus();

			if (this.options.rtspMode) {
				if (this.fbvlc) {
					this.video[0].stop();
					this.video[0].playlist.items.clear();
				} else {
					this.video[0].playlist.stop();
					this.video[0].playlist.clear();
				}
			}
		},
		
		// Palauttaa kuvan koordinaatit canvasissa
		_top: function () {
			if (this.image) {
				var top = (this.options.height - this.image.height) / 2;
				if (top > 0) return top;
			}
			return 0;
		},
		_bottom: function () {
			if (this.image) {
				var bottom = this._top() + this.image.height;
				if (bottom < this.options.height) return bottom;
			}
			return this.options.height;
		},
		_left: function () {
			if (this.image) {
				var left = (this.options.width - this.image.width) / 2;
				if (left > 0) return left;
			}
			return 0;
		},
		_right: function () {
			if (this.image) {
				var right = this._left() + this.image.width;
				if (right < this.options.width) return right;
			}
			return this.options.width;
		},
		_width: function () { return this.image ? this.image.width : this.width(); },
		
		_clearContext: function (context) {
			context.save();
			context.setTransform(1, 0, 0, 1, 0, 0);
			context.clearRect(0, 0, this.options.width, this.options.height);
			context.restore();
		},

		_draw: function () {
			if (this.canvas) {
				this._clearContext(this.canvas[0].getContext("2d"));
				if (this.options.error.length > 0) {
					this._drawError();
				} else if (!this.theGameIsOn && this.image) {
					this._drawAudioStatus();
					this._drawStatus();
				}
			}
		},

		_drawAudioStatus: function () {
			if (!this._shouldShowAudioStatus() || !this.audioStatus) return;

			var status = this.audioStatus;
			if (status === null) return;

			var spectrum = status.spectrum.split(",");
			var scaledSpectrum = scaleSpectrum(spectrum, this._width());
			
			var top = this._top();
			var bottom = this._bottom();
			var left = this._left();

			var freqsPerHorizontalPixel = frequenciesPerHorizontalPixel(spectrum, this._width());
			var low = status.low / freqsPerHorizontalPixel;
			var high = status.high / freqsPerHorizontalPixel;
			
			var ctx = this.canvas[0].getContext("2d");
			
			var green = "rgba(0, 255, 0, 0.8)";
			var red = "rgba(255, 0, 0, 0.8)";
			
			ctx.fillStyle = green;
			for (var i = 0; i < this._width(); ++i) {
				var s = scaledSpectrum[i] / 500;

				if (i > low && i < high && s > status.sensitivity) {
					ctx.fillRect(left + i, bottom - status.sensitivity, 1, status.sensitivity);
					ctx.fillStyle = red;
					ctx.fillRect(left + i, bottom - s, 1, s - status.sensitivity);
					ctx.fillStyle = green;
				} else {
					ctx.fillRect(left + i, bottom - s, 1, s);
				}
			}
			
			ctx.fillStyle = "rgba(255, 255, 0, 1)";
			ctx.fillRect(left + low, top, 2, this.height());
			ctx.fillRect(left + high, top, 2, this.height());
			
			ctx.fillRect(left, bottom - status.sensitivity, this._width(), 2);
		},
		
		_drawStatus: function () {
			if (!this._shouldShowStatus() || this.status.length === 0) return;
			
			var left = this._left();
			var top = this._top();
			var bottom = this._bottom();
			
			var context = this.canvas[0].getContext("2d");
			
			var lines = this.status.split("\n");
			context.font = this.font;
			
			var maxLineWidth = 0;
			for (var i = 0; i < lines.length; ++i) maxLineWidth = Math.max(context.measureText(lines[i]).width, maxLineWidth);

			// Piirretään tausta..
			context.fillStyle = this.backgroundColor;
			context.fillRect(left, top, maxLineWidth + this.defaultMargin * 2, (lines.length + 1) * (this.fontSize + this.defaultMargin) + this.defaultMargin);
			
			// Piirretään tallennuspallo
			var r = 3;
			context.fillStyle = this.recordingStatus ? "rgba(255, 84, 84, 1.0)" : "rgba(255, 255, 255, 0.4)";
			context.arc(left + this.defaultMargin + r, top + this.defaultMargin + (this.fontSize / 2), r, 0, 2 * Math.PI);
			context.fill();
			top += this.fontSize;

			// ..ja status-rivit
			context.fillStyle = this.textColor;
			for (var i = 0; i < lines.length; ++i) context.fillText(lines[i], left + this.defaultMargin, top += this.defaultMargin + this.fontSize);
		},
		
		_drawError: function () {
			// Piirretään tausta..
			var context = this.canvas[0].getContext("2d");
			context.fillStyle = "rgba(0, 0, 0, 0.7)";
			context.fillRect(0, 0, this.width(), this.height());

			var errorImageY = this.height() / 2 - this.errorImage.height / 2;
			context.drawImage(this.errorImage, this.width() / 2 - this.errorImage.width / 2, errorImageY);
			
			context.fillStyle = this.textColor;
			context.font = this.font;
			var textSize = context.measureText(this.options.error);
			context.fillText(this.options.error, this.width() / 2 - textSize.width / 2, errorImageY + this.errorImage.height + 30);
		},
		
		stopGame: function () {
			this.theGameIsOn = false;
		},
		
		startGame: function () {
			this.theGameIsOn = true;
			this._draw();
		},
		
		_animate: function () {
			window.requestAnimFrame = (function (callback) {
				return window.requestAnimationFrame || window.webkitRequestAnimationFrame || window.mozRequestAnimationFrame || window.oRequestAnimationFrame || window.msRequestAnimationFrame ||
				function (callback) {
					window.setTimeout(callback, 1000 / 60);
				};
			})();
			var that = this;
			
			var wormPos;
			var wormLastUpdate = 0;
			var wormSpeed;
			var wormLength;
			var wormParts = [];
			var wormCaster = new THREE.Raycaster();
			var appleMesh = new THREE.Mesh(
				new THREE.BoxGeometry(that.brushWidth, that.brushHeight, that.brushWidth),
				new THREE.MeshBasicMaterial({ color: 0xff0000, transparent: true, opacity: 0.6 }));
			appleMesh.visible = false;
			that.scene.add(appleMesh);
			
			function moveAppleToRandomPosition() {
				do {
					var c = that.frameToScene(parseInt(Math.random() * that.options.width), parseInt(Math.random() * that.options.height));
					appleMesh.position.set(c.x, c.y, 0);
				} while (wormCaster.intersectObjects(wormParts).length > 0 || wormCaster.intersectObjects(that.brushPointObjects).length > 0)
			};
			
			function initWorm() {
				wormPos = new THREE.Vector2(100, 100);
				wormSpeed = 200;
				wormLength = 3;
				for (var i = 0; i < wormParts.length; ++i) that.scene.remove(wormParts[i]);
				wormParts = [];
			};
			initWorm();
			
			function render() {
				if (that.theGameIsOn && wormLastUpdate + wormSpeed < (new Date).getTime()) {
					if (!appleMesh.visible) appleMesh.visible = true;
				
					var p = wormPos;				
					if (that.wormDirection === 1) p.y -= that.brushHeight - 1;
					else if (that.wormDirection === 2) p.x += that.brushWidth + 1;
					else if (that.wormDirection === 3) p.y += that.brushHeight + 1;
					else if (that.wormDirection === 0) p.x -= that.brushWidth - 1;
					p = that._toBrushPosition(p);

					wormCaster.setFromCamera(that.frameToNDC(p.x + that.brushWidth / 2, p.y + that.brushHeight / 2), that.camera);
					var intersects = wormCaster.intersectObject(appleMesh);
					if (intersects.length > 0) {
						wormLength++;
						moveAppleToRandomPosition();
					}
					if (wormCaster.intersectObjects(wormParts).length > 0 || wormCaster.intersectObjects(that.brushPointObjects).length > 0) {
						initWorm();
					} else {
						wormParts.push(that._drawBrushPoint(p));
						if (wormParts.length > wormLength) that.scene.remove(wormParts.shift());
						wormPos = p;
						wormLastUpdate = (new Date).getTime();
						if (wormSpeed > 50) wormSpeed--;
					}
				}
			
				if (that.imageUpdated) {
					that.imageUpdated = false;
					
					if (that.frameMaterial.map) that.frameMaterial.map.dispose();
										
					var texture = new THREE.Texture(that.image);
					texture.minFilter = THREE.NearestFilter;
					that.frameMaterial.map = texture;
					that.frameMaterial.needsUpdate = true;
					that.frameMaterial.map.needsUpdate = true;
					that.frameMaterial.DoubleSide = true;
					
					if(that.options.rotation == 90 ||that.options.rotation == 270) {
						var maxWidth = that.options.height;
						var maxHeight = that.options.width;
					} else {
						var maxWidth = that.options.width;
						var maxHeight = that.options.height;
					}
					var w = that.image.width;
					var h = that.image.height;
					var ratio = w / h;

					var realW = w;
					var realH = h;
					
					if (w > maxWidth){
						ratio = maxWidth / w;
						realW = maxWidth;
						realH = h * ratio
						h = h * ratio;
						w = w * ratio;
					}

					if (h > maxHeight){
						ratio = maxHeight / h;
						realH = maxHeight;
						realW = w * ratio;
						w = w * ratio;
						h = h * ratio;
					}
					
					w = parseInt(realW);
					h = parseInt(realH);

					var rebuild = that.frameMesh.geometry.parameters.width !== w || that.frameMesh.geometry.parameters.height !== h
					if(that.options.flipChanged || that.options.rotationChanged)  {
						rebuild = true;
						that.options.flipChanged = false;
						that.options.rotationChanged = false;
					}

					if (rebuild) {
						var m = that.mask();
						
						that.frameMesh.geometry = new THREE.BoxGeometry(w, h, 0);
						that.frameMesh.needsUpdate = true;

						if(that.options.rotation == 90 || that.options.rotation == 270) {
							that.renderer.setSize(h, w);
							$(that.renderer.domElement).css("left", (that.options.width - h) / 2);
							$(that.renderer.domElement).css("top", (that.options.height - w) / 2);			
							that.camera
							that.camera.left = h / -2;
							that.camera.right = h / 2;
							that.camera.top = w / 2;
							that.camera.bottom = w / -2;
							that.camera.updateProjectionMatrix();
						} else {
							that.renderer.setSize(w, h);
							$(that.renderer.domElement).css("left", (that.options.width - w) / 2);
							$(that.renderer.domElement).css("top", (that.options.height - h) / 2);			
							that.camera
							that.camera.left = w / -2;
							that.camera.right = w / 2;
							that.camera.top = h / 2;
							that.camera.bottom = h / -2;
							that.camera.updateProjectionMatrix();
						}
						that._updateBrushSize();
						that._updateMotionPointSize();
						that._updateAreas();
						that.setMask(m);
					}
					that.frameMesh.scale.set( that.options.fliph ? -1 : 1, that.options.flipv ? -1 : 1 , 1);
					that.frameMesh.rotation.set(0, 0, -that.options.rotation * Math.PI/180);
				}

				that.renderer.render(that.scene, that.camera);
				requestAnimationFrame(render);
			};	
			render();
		},
		
		_clearAreas: function () {
			for (var i = 0; i < this.areaObjects.length; ++i) {
				var area = this.areaObjects[i];
				for (var key in area) {
					var object = area[key];
					if (object instanceof Array) for (var j = 0; j < object.length; ++j) this.scene.remove(object[j]);
					else this.scene.remove(object);
				}
			}
			this.areaObjects = [];
		},
		
		_relToAbs: function (point) { // pointti on taulukko jossa x ja y
			return [point[0] * this.renderer.domElement.width - this.renderer.domElement.width * 0.5, (1 - point[1]) * this.renderer.domElement.height - this.renderer.domElement.height * 0.5];
		},

		_absToRel: function (point) { // pointilla on x- ja y-jäsenmuuttujat
			return [(point.x + this.renderer.domElement.width * 0.5) / this.renderer.domElement.width, 1 - (point.y + this.renderer.domElement.height * 0.5) / this.renderer.domElement.height];
		},
		
		_drawVertex: function (point, isSelected) { // point on taulukko jossa suhteelliset x:n ja y:n arvot
			var p = this._relToAbs(point);
				
			// Tehdään esin iso pallo josta jää reunat näkyviin, kun pieni pallo laitetaan päälle
			var edge = new THREE.CircleGeometry(6, 12);
			var edgeMesh = new THREE.Mesh(edge, new THREE.MeshBasicMaterial( { color: 0xffffff, transparent: true, opacity: isSelected ? 0.8 : 0.4 }));
			edgeMesh.position.x = p[0];
			edgeMesh.position.y = p[1];
			
			// Pieni pallo eli sisäosa
			var innerArea = new THREE.CircleGeometry(4, 12);
			var innerAreaMesh = new THREE.Mesh(innerArea, new THREE.MeshBasicMaterial( { color: 0x3367d5, transparent: true, opacity: isSelected ? 0.8 : 0.4 }));
			edgeMesh.add(innerAreaMesh);
			
			return edgeMesh;
		},
		
		_drawClosedArea: function (areaPoints, isSelected) {
			var area = new Map();
			area["type"] = "closed";
			if (areaPoints.length > 1) {
				// Tehdään Shape-objekti alueen pisteiden perusteella
				var shape = new THREE.Shape();
				var p = this._relToAbs(areaPoints[0]);
				shape.moveTo(p[0], p[1]);
				for (var i = 1; i < areaPoints.length; ++i) {
					p = this._relToAbs(areaPoints[i]);
					shape.lineTo(p[0], p[1]);
				}

				// Tehdään Geometry-objekti shapen perusteella
				var areaGeom = shape.makeGeometry();
				this.areaMesh = new THREE.Mesh(areaGeom, new THREE.MeshBasicMaterial( { color: 0x3367d5, transparent: true, opacity: isSelected ? 0.5 : 0.2 } ));
				this.scene.add(this.areaMesh);
				area["area"] = this.areaMesh;

				// Tehdään alueelle reunat
				var edgeMaterial = new THREE.LineBasicMaterial({ color: 0x007dd6, transparent: true, opacity: isSelected ? 0.8 : 0.4 });
				var edges = [];
				var points = shape.extractPoints().shape;
				for (var i = 0; i < points.length; ++i) {
					var edge = new THREE.Geometry();
					edge.vertices.push(new THREE.Vector3(points[i].x, points[i].y, 0));
					
					// Jos ollaan viimeisessä pisteessä, loppupisteeksi ensimmäinen piste
					if (i === points.length - 1) edge.vertices.push(new THREE.Vector3(points[0].x, points[0].y, 0));
					else edge.vertices.push(new THREE.Vector3(points[i + 1].x, points[i + 1].y, 0));
					
					var line = new THREE.Line(edge, edgeMaterial);
					this.scene.add(line);
					edges.push(line);
				}
				area["areaEdges"] = edges;
			}

			// Tehdään alueen kärkiin pallot
			var nodes = [];
			for (var i = 0; i < areaPoints.length; ++i) {
				var mesh = this._drawVertex(areaPoints[i], isSelected);
				this.scene.add(mesh);
				nodes.push(mesh);
			}
			area["nodes"] = nodes;
			return area;
		},
		
		_drawLine: function (areaPoints, isSelected) {
			var area = new Map();
			area["type"] = "line";
			if (areaPoints.length < 2) return area;
			
			// Tehdään viivat
			var edgeMaterial = new THREE.LineBasicMaterial({ color: 0x007dd6 });
			var edges = [];
			for (var i = 0; i < areaPoints.length - 1; ++i) {
				var start = this._relToAbs(areaPoints[i]);
				var end = this._relToAbs(areaPoints[i + 1]);
				
				var edge = new THREE.Geometry();				
				edge.vertices.push(new THREE.Vector3(start[0], start[1], 0));
				edge.vertices.push(new THREE.Vector3(end[0], end[1], 0));
				
				var line = new THREE.Line(edge, edgeMaterial);
				this.scene.add(line);
				edges.push(line);
			}
			area["areaEdges"] = edges;

			// Tehdään alueen kärkiin pallot
			var nodes = [];
			for (var i = 0; i < areaPoints.length; ++i) {
				var mesh = this._drawVertex(areaPoints[i], isSelected);
				this.scene.add(mesh);
				nodes.push(mesh);
			}
			area["nodes"] = nodes;
			return area;
		},
		
		_updateBrushSize: function () {
			if(this.options.rotation == 90 || this.options.rotation == 270) {
				this.brushWidth = this.frameMesh.geometry.parameters.height / 36;
				this.brushHeight = this.frameMesh.geometry.parameters.width / 44;
			} else {
				this.brushWidth = this.frameMesh.geometry.parameters.width / 44;
				this.brushHeight = this.frameMesh.geometry.parameters.height / 36;
			}
			this.brushGeom = new THREE.BoxGeometry(this.brushWidth + 1, this.brushHeight + 1, 0);
		},
		
		// @point: koordinaatti kamera-alueen koordinaatistossa
		_toBrushPosition: function (point) {
			return new THREE.Vector2(parseInt(point.x / this.brushWidth) * this.brushWidth, parseInt(point.y / this.brushHeight) * this.brushHeight);
		},

		// @point: Piirrettävän maalialueen vasemman yläkulman koordinaatti kamera-alueen koordinaatistossa.
		_drawBrushPoint: function (point) {
			var mesh = new THREE.Mesh(this.brushGeom, this.brushMaterial);
			var scenePoint = this.frameToScene(point.x, point.y);
			mesh.position.set(scenePoint.x + this.brushWidth / 2, scenePoint.y - this.brushHeight / 2, 0);
			this.scene.add(mesh);
			return mesh;
		},
		
		_clearBrushPoints: function () {
			for (var i = 0; i < this.brushPointObjects.length; ++i) this.scene.remove(this.brushPointObjects[i]);
			this.brushPointObjects = [];
		},
		
		_showBrushPoints: function (state) {
			for (var i = 0; i < this.brushPointObjects.length; ++i) this.brushPointObjects[i].visible = state;
		},

		_updateMotionPointSize: function () {
			if(this.options.rotation == 90 || this.options.rotation == 270) {
				this.motionPointWidth = this.frameMesh.geometry.parameters.height / 36;
				this.motionPointHeight = this.frameMesh.geometry.parameters.width / 44;
			} else {
				this.motionPointWidth = this.frameMesh.geometry.parameters.width / 44;
				this.motionPointHeight = this.frameMesh.geometry.parameters.height / 36;
			}
			this.motionPointGeom = new THREE.BoxGeometry(this.motionPointWidth + 1, this.motionPointHeight + 1, 0);
		},

		_drawMotionPoint: function (point, value) {
			var mesh = new THREE.Mesh(this.motionPointGeom, new THREE.MeshBasicMaterial({ color: "#00ff00", transparent: true, opacity: Math.max(0.6 * (value / 255), 0.1)}));
			var scenePoint = this.frameToScene(point.x, point.y);
			mesh.position.set(scenePoint.x + this.motionPointWidth / 2, scenePoint.y - this.motionPointHeight / 2, 0);
			this.scene.add(mesh);
			return mesh;
		},

		_clearMotionPoints: function () {
			for (var i = 0; i < this.motionPointObjects.length; ++i) this.scene.remove(this.motionPointObjects[i]);
			this.motionPointObjects = [];
		},

		_updateAreas: function () {
			this._clearAreas();
			for (var j = 0; j < this._areas.length; ++j) {
				var areaPoints = this._areas[j]["points"];
				var type = this._areas[j]["type"];
				if (type === "closed") {
					this.areaObjects.push(this._drawClosedArea(areaPoints, this._areas[j]["id"] === this.options.selectedArea));
				} else if (type === "line") {
					this.areaObjects.push(this._drawLine(areaPoints, this._areas[j]["id"] === this.options.selectedArea));
				}
			}
		}
	});

})(jQuery);
